#include "ArnholdMesh.h"
#include "LedPanel.h"
#include <Arduino.h>

#define   STATION_SSID     "lemur"
#define   STATION_PASSWORD ""

const uint32_t BLIMP_ID = 88789821;

// Broker options:
  // test.mosquitto.org
  // Sud's broker: 172.20.10.10 when connected to
  // his hotspot.
#define   MQTT_BROKER      ""
#define   MQTT_TOPIC       ""


ArnholdMesh thisNode;
LedPanel panel(32, 32);

void setup() {
  Serial.begin(115200);
  panel.beginPanel();
  thisNode.setPanel(&panel);
  thisNode.setStationInfo(STATION_SSID, STATION_PASSWORD);
  thisNode.setMQTTInfo(MQTT_BROKER, MQTT_TOPIC);
  thisNode.amIMQTTBridge(false);
  thisNode.init();
}
void loop() {
  thisNode.update();
//  if (Serial.available() < 2){
//    //do nothing
//    thisNode.update();
//  } else {
//    String command = Serial.readString();
//    Serial.print("the command is: ");
//    Serial.println(command);
//    thisNode.sendOverAllFreqs(BLIMP_ID, command);
//  }
}

void interpretDashInstructions(String& msg){
  // Deserialize different dash topics as they are received
  String result;
  std::list<uint32_t> nodesToSend;
  Serial.println(msg);
  // If dash topic is Command
  if(thisNode.dashDeserialize(msg, "Command", result, nodesToSend)){
    // Interpret what command on the dash means
    char numChar = result[0];
    String commandContent = result.substring(1);
    if(numChar == 'G'){
      String topo = thisNode.mesh.subConnectionJson();
      thisNode.sendJsonToDash("Debug", topo);
      return;
    }
    else {
      thisNode.sendOverAllFreqs(nodesToSend, result);
    }
  }
  // If the dash topic is MasterJoystick
  else if(thisNode.dashDeserialize(msg, "MasterJoystick", result, nodesToSend)){
     // Interpret what a dash MasterJoystick message means
     String joystickcontrol = "J";
     joystickcontrol += result;
     thisNode.sendOverAllFreqs(nodesToSend, result);
     return;
  }
}

void vehicleInstruction(char instructionType, String instrData, uint32_t from){
  // Do something with the data receieved over the network
  // Based on what the instructionType is, do something different with the instrData
  // instructionType CAN NOT BE:
  // 'B', 'O', 'g', 'm', 'u', 'R', 'D'
  //  I am using those within ArnholdMesh for testing purposes
  switch(instructionType){
     default:
      break;
  }
}

// Callback function needed for painlessMesh
void receivedCallback(uint32_t from, String &msg) {
  Serial.printf("startHere: Received from %u msg=%s\n", from, msg.c_str());
  thisNode.lastMsgReceieved = millis();
  thisNode.colorsL[2] = 5;
  thisNode.panel->topLeftQuadrant(thisNode.colorsL[0], thisNode.colorsL[1], thisNode.colorsL[2]);
  thisNode.interpretInstruction(from, msg);
}


// Callback function needed for MQTT
void mqttCallback(char* topic, uint8_t* payload, unsigned int length) {
  char* cleanPayload = (char*)malloc(length+1);
  memcpy(cleanPayload, payload, length);
  cleanPayload[length] = '\0';
  String msg = String(cleanPayload); 
  free(cleanPayload);
  Serial.println(msg);
  interpretDashInstructions(msg);
}

void newConnectionCallback(uint32_t nodeId) {
  Serial.printf("--> startHere: New Connection, nodeId = %u\n", nodeId);
  digitalWrite(LED_BUILTIN, HIGH);
  if(thisNode.bridgeId == 0){
      String askpayload = "g";
      thisNode.sendOverAllFreqs(nodeId, askpayload);
  }
}
