**Repo is Currently a Work in Progress**
I will try to get more details into it so that it will be easier to build but for now, here are the things that I have not included (or not include a completed documentation):

1. ways to put the electronic component on the mount
2. pullup resistor circuit between OpenMV and the ESP32 (though I think the diagram should suffice)
3. a sample remote controller code to vary the parameter of the blimp

Hopefully, I can get all this documented by 8/30/2021

**What to Expect**
As of this point, we have done the following contribution towards making our robots ready for the November competition:
1. Set a specific threshold of colors for detection (by default green)
2. Rotate horizontally when the blimp do not see any targeted blob
3. Move towards the targeted blobs when it sees it, while keeping the center of the blob at the center of the camera frame 

Here's a link to a blogpost showing the video of how it works: https://uclalemur.com/blog/pid-control-for-blimp-motion

**Components Needed**
1. One ESP32 Feather Board, Stacking Header (https://www.adafruit.com/product/3405)
2. One Adafruit Neopixel LED Panel, Plain Header (https://www.adafruit.com/product/2945)
3. One Adafruit Radio FeatherWing 433MHz, Stacking Header (https://www.adafruit.com/product/3230)
4. Wires for I2C and motor connection
5. Electronic Components Mount* (in General 3D Print Design)
6. 1 to 3 Lipo Batteries 3.7V-3.8V (https://www.amazon.com/dp/B07L9SHHFX)
7. One OpenMV H7 or H7 Plus (https://openmv.io/)
8. Velcro Tape**
9. Two Blimp*** (https://www.amazon.com/dp/B005FYCBR6)
10. Two 2.2k Ohms resistors
11. Helium
12. Three DC motors (https://www.amazon.com/dp/B076M7G24G)
13. Three Propellers (https://www.amazon.com/dp/B076M7G24G)
14. Putty (https://www.amazon.com/dp/B08DTHJTSZ)


*: Although we are currently using a 3D-printed gondola as a mount, anything else that can hold the components together can be used.

**: velcros are used to hold the mount to the blimp. If you don't have it, normal tape would suffice but it will be a little bit messier

***: The total mass of all components weight approximately around 120g. Therefore, a blimp that can carry that much weight is needed. For the Amazon product listed, you need two of them for the blimp to float in the air.

**Connecting the ESP32 Feather with OpenMV**

1. If you are planning on using the I2C communication protocol, you need to connect the following from Feather to OpenMV:
    * GND pin - GND pin
    * SCL/22 pin - P4 pin
    * SDA/23 pin - P5 pin
2. Just like the ESP32, we need to make sure the battery connected to OpenMV is of the right polarity. With the battery suggested, you need to reverse the wiring in order to connect a battery to OpenMV.
4. If the battery is properly connected, the OpenMV LED will blink white at the start. In addition to that, a debugging feature was added such that the OpenMV LED will be red once it is switched on. Everytime a function is called, it will blink green for the period of the function. 
3. In addition to that, since OpenMV is not a stackable feather component, you need to add pull up resistors for both the SDA and SCL lines. The way to do this is by connecting two 2.2kOhms resistors where one end is connected to ESP32 3V and the other one is connected to the SDA and SCL respectively. 

Pullup resistor circuit example:

![Pullup Resistor Setup](/Pictures/Pullup_Resistor.png)

(https://www.analog.com/en/technical-articles/i2c-primer-what-is-i2c-part-1.html)

Camera setup example:
![Camera Setup](/Pictures/Bllimp_Camera_Setup.jpg)

Note: Failure to properly setup the I2C protocol might cause the code to not work properly

**Connecting the ESP32 Feather with the Motor Shield**

1. The Feather and the shield should be stackable. Make sure you have soldered the pins properly.
2. The shield runs on a separate power source. Make sure that you have a 3.7V Lipo Battery if you are planning to use the shield.
3. If the battery is properly connected, a green LED should light up. This LED also approximately indicate how much power the battery has left as it will get dimmer with lower voltage.
4. Connect the vertical motor to M1, right motor to M2, and left motor to M4.
5. The motor itself can be connected  differently if you are using a different vehicle. If so, changes might be needed for the motor controller code section.

Note: depending on the wiring and the propeller design, the propeller might push the blimp in the opposite direction as the code intended. Ways to solve this problem include reversing the polarity wiring of the motor, use the oppsite-direction propeller of the one you have right now (if it is clockwise, use the counter-clockwise propeller instead), or change the code so that all BACKWARD motor calls are FORWARD and vice versa.

DC Motor Shield setup example:
![Blimp Setup](/Pictures/Blimp_Motor_Setup.jpg)

**Other Physical Setup**

1. The stack, starting from the bottom is the ESP32 Feather, Radiowing Feather, DC Motor Shield, and the Neopixel LED Panel.
2. Just like the OpenMV, reverse the wiring before connecting a battery to the ESP32. Optional: if weight is an issue, you can use a single battery and connect the ESP32 and OpenMV in parallel. However, it is strongly advised that you use a separate battery for the DC Shield as it consumes a lot of power.
3. Fill the two blimp balloons with Helium until both of them together can carry around 130-140g. In order to test this, tie a string to the blimp and attach the string to a heavier object like a smartphone. First, weight the smartphone by itself. Then, weight the smartphone when it is attached to the string. The difference will be what the blimp can carry.
4. Using the velro tape, attach the top of the electronic mount to the bottom of the blimp
5. Similarly, use velcro tape to attach the OpenMV camera to the front of the mount in such a way that the camere is facing the forward direction.
6. Plug the motor to its corresponding location on the mount
7. Place the ESP32 and the rest of the stack into the mount
8. Test to see if the whole blimp floats weightlessly in the air. If it is falling down, add more helium to the blimp balloon. If it is floating up, add putty into the mount. Repeat this until it approximately stays still in the air without any external force acting on it. 

Blimp stack setup:
![Blimp Stack](/Pictures/Blimp_Stack_Setup.jpg)

Blimp front view:
![Blimp Front View](/Pictures/Blimp_Front_View.jpg)

Blimp bottom view: 
![Blimp Bottom View](/Pictures/Blimp_Bottom_View.jpg)

Complete setup:
![Blimp complete](/Pictures/Blimp_Complete.jpg)

**Setting Up the General Libraries and Codes**

1. Go to https://git.uclalemur.com/arnhold/espmeshnetworking/-/tree/master/MeshCode
2. In that repository, follow the instructions to install the ArnholdMesh.zip and Painless_Mesh.zip libraries. 
3. If you are planning on doing demo using the LED panel, please also install Adafruit Neopixel library through the Library Manager in your Arduino IDE
4. If you are using the Adafruit Motor shield, **do not** install the default library through the library manager. Instead, in your Arduino IDE, go to Sketch > Include Library > Add .ZIP Library
5. Download or Clone the Adafruit_Motor_Shield_V2_Library.zip from this repository and use that for Step 4. The reason we are using an alternate library from the default one is because the motor library and the mesh library uses the same name for one of their macros. This changes the macro's name in the motor library to ensure that this doesn't happen.
6. Download the PID library, which is needed for the motion control. (https://playground.arduino.cc/Code/PIDLibrary/)

**Running the Code**

1. Create a folder name attacker_main in your directory of choice.
2. Copy everthing in the Main Code repository into the attacker_main folder.
3. Open feather_main_personaldemo.ino and upload the sketch into your ESP32 Feather
4. Upload the OpenMV_main.py into your OpenMV camera as main.py so that it can run without being connected to the IDE
5. By default, it should start detecting and responding to green blob.

**Varying the Parameters**

If we want to vary the behaviour of the blimp, there are a couple of parameter that we can use to do that. The section below explains what message needs to reach the ESP32 through the mesh network. However, the methodology of how the message is sent to the blimp node is completely up to you. Another section further below will give you an example remote controller you can use (although not the best)

1. Color Detection Threshold (C)
    * By default, the blimp will start trying to detect the biggest green blob. However, there are 2 ways we can change the threshold so that the program will look for other colored blobs.
    * The first way is to provide a single-digit color code. The code along with the command character:
        - "C1": generic green blob
        - "C2": generic blue blob
        - "C5": generic red blob
    * If the generic colors are not enough, you can also prescribe a specific thresholds for the robot to use. The format will be something like: 
        - "C(Lmin) (Lmax) (Amin) (Amax) (Bmin) (Bmax)", without the brackets
        - "C30 100 12 -19 0 -120"
    

2. PID Constants (T)
    * You can also vary the PID constant for the horizontal motion. By default, these are the constants:
        - Kpx = 1
        - Kix = 0.05
        - Kdx = 0.25
    * In order to change the constant which are double datatype, the generic form of the message should be:
        - "TP0.1 I0.2 D0.3"
    * For this command, you can also just change one or 2 parameters by doing the command like:
        - "TP0.5"
        - "TI0.9 D1.5"

3. Pause the Blimp (P)
    * If you want to stop the mesh from moving, a quick way is to use the pause feature:
        - "P1": pause the blimp program
        - "P0": resume the blimp program

4. Change the Base Speed (S)
    * The base speed is used to move towards the tracked object. If it is 0, it the blimp will not actively move towards the object if it is detected by the camera. The command format is:
        - "S0"
        - "S250"

**LED Panel Debugging with the Blimp**

- OpenMV LED: blink green when a function is called. red on standby
- PIN31: state of the blimp. red is standby, yellow is paused, white is seeking, blue is done
- PIN30: kp change. blink green if new kp is higher then old kp. red if less. white if the same
- PIN29: ki change. blink green if new ki is higher then old ki. red if less. white if the same.
- PIN28: kd change. blink green if new kd is higher then old kd. red if less. white if the same
- PIN27: base speed indicator. the white LED intensity scale with the base speed, with 255 speed being the brightest
- PIN26: minimum threshold for color detection, display the minimum LAB value in terms of RGB 
- PIN25: maximum threshold for color detection, display the maximum LAB value in terms of RGB
- PIN24: left motor speed. display the speed of the motor that scales with the light intensity. led will be red if the motor is pushing the blimp forward, blue if backward
- PIN17: vertical motor speed. display the speed of the motor that scales with the light intensity. led will be red if the motor is pushing the blimp upward, blue if downward
- PIN16: right motor speed. display the speed of the motor that scales with the light intensity. led will be red if the motor is pushing the blimp forward, blue if backward

LED Panel numbering:

![LED Panel](/Pictures/LEDPanel.png)

**Using the Camera.cpp Class Separately**

**Primitive Remote Control**

One of our end goal is to integrate a dashboard to do high level control for the blimps. While we are still working on that front, we created a primitive remote controller using the Arduino Serial Monitor from another ESP32 to control a blimp. Here's how to do it:
1. Using another ESP32, stack a Neopixel LED Panel on top of it for debugging purposes
2. Connect the ESP32 to a computer using a USB cable
3. Go to the "Remote Controller" directory in "Codes" folder and open the "MainRemoteController.ino" file in your Arduino IDE.
4. Before you start using it, there is one thing you need to do which is to change the ID of the blimp you want to receive command. The line of code in the file is:

        const uint32_t BLIMP_ID = 88789821;

5. Change the ID correponding to your ESP32 node ID that you are using on the blimp
6. If all are connected properly, open up your Serial Monitor and press G. This should display the mesh network topology and it should show that the controller and the blimp are connected.
7. In order to send a command, you need to include the blimp position in the mesh topology in addition to the command in the format "Node Position + Message". For example, if your node position is 0, the string you need to send to change the color threshold to red would be "0C5" in the Serial Monitor. If your node position is 1, then the message you want to send when you want to change the P and I constant values will be "1TP0.2 I0.25"


1. Install OpenMV Arduino RPC Library into your Arduino IDE.
2. At the top of your main code, include the following lines to create the i2c interface that the OpenMV will communicate with the feather:

        #include "Camera.h"
        openmv::rpc_scratch_buffer<256> scratch_buffer; // All RPC objects share this buffer.
        openmv::rpc_i2c_master interface(0x12, 100000); //to make this more robust, consider making address and rate as constructor argument
3. You need to pass the interface's address when you are constructing the camera:
        
        Camera cam(&interface);
4. Now, you should be able to use all the function in the Camera.h file.
5. If you want to make a new function in the OpenMV and create an interfacing function corresponding to it on Arduino, please refer the RPC library for further clarification. https://github.com/openmv/openmv-arduino-rpc

