//TASK:
// - Scale this function so that it can command as many cars to specified targets

#include "Camera.h"
#include "ArnholdMesh.h"
#include "LedPanel.h"
#include <Wire.h>
#include <string>

#define DDRIVE_MIN -160 //The minimum value x or y can be.
#define DDRIVE_MAX 160  //The maximum value x or y can be. In this case, it will be the width of camera fram
#define MOTOR_MIN_PWM -255 //The minimum value the motor output can be.
#define MOTOR_MAX_PWM 255 //The maximum value the motor output can be.

#define   STATION_SSID     "lemur"
#define   STATION_PASSWORD "lemur9473"
#define   MQTT_BROKER      "test.mosquitto.org"
#define   MQTT_TOPIC       "LEMUR"

//Maybe we can do just 50 pixel or higher, maximum speed or something like that

int targetID = 0;
int vehicleID = 2;
int targetID2 = 7;
int vehicleID2 = 8;

const int BAUD_RATE = 115200;
int completionCounter = 20;
int completionCounter2 = 20;
unsigned long time_now = 0;
uint32_t car_nodeId = 541404797; //RED CAR;
uint32_t car2_nodeId = 88790569; //BLUE CAR

//Create the interface that will be used by the camera
openmv::rpc_scratch_buffer<256> scratch_buffer; // All RPC objects share this buffer.
openmv::rpc_i2c_master interface(0x12, 100000); //to make this more robust, consider making address and rate as constructor argument

ArnholdMesh thisNode;
Camera cam(&interface);
LedPanel panel(NUMPIXELS,PIN_PIXELS);

void setup() {
  Serial.begin(BAUD_RATE);
  Wire.begin();
  interface.begin();
  panel.beginPanel();
  panel.fullPanelLight(10, 0 , 0);
  time_now = millis();
  while(millis() < time_now + 5000){
    //wait approx. 5000 ms
    Serial.print("time");
  }
  //delay(5000);
  panel.fullPanelLight(5, 5, 0);
  thisNode.setStationInfo(STATION_SSID, STATION_PASSWORD);
  thisNode.setMQTTInfo(MQTT_BROKER, MQTT_TOPIC);
  thisNode.amIMQTTBridge(false);
  thisNode.init();
  Serial.println("setup done");
}

void loop() {
  time_now = millis();
  while(millis() < time_now + 10){
    //wait approx. 5000 ms
    thisNode.update();
  }

  thisNode.update();
  //Serial.print("completionCounter: ");
  //Serial.println(completionCounter);
  if (completionCounter == 0 && completionCounter2 == 0){
    String vehicleInput = "M";
    vehicleInput += 0;
    vehicleInput += "a";
    vehicleInput += 0; 
    while(1){
      panel.fullPanelLight(0, 10 , 0);
      thisNode.sendOverAllFreqs(car_nodeId, vehicleInput);
      thisNode.sendOverAllFreqs(car2_nodeId, vehicleInput);
      while(millis() < time_now + 100){
        //wait approx. 5000 ms
      thisNode.update();
      }
    }
  }
  int target_cx = 0; 
  int target_cy = 0;
  int target_angle = 0; 
  int vehicle_cx = 0; 
  int vehicle_cy = 0;
  int vehicle_angle = 0; 
  bool notDone = cam.exe_apriltag_detection(targetID, &target_cx, &target_cy, &target_angle);
  //Serial.println(notDone);

  //CAR 1
  if (notDone == false){
    String w = "M0a0";
    thisNode.sendOverAllFreqs(car_nodeId, w);
    if (completionCounter > 0){
          completionCounter--;
    }
  } else {
    completionCounter = 20;
    int x_vec = 0;
    int y_vec = 0;
    if (cam.exe_apriltag_detection(vehicleID, &vehicle_cx, &vehicle_cy, &vehicle_angle)){
      cartesianVectorGenerator(vehicle_cx, vehicle_cy, target_cx , target_cy, vehicle_angle, x_vec, y_vec);
      int leftMotor = 0;
      int rightMotor = 0;
//    Serial.print("x vec: ");
//    Serial.println(x_vec);
//    Serial.print("y vec: ");
//    Serial.println(y_vec);
      actuatorOutputGenerator(x_vec, y_vec, leftMotor, rightMotor);
      String vehicleInput = "M";
      vehicleInput += leftMotor; 
      vehicleInput += "a";
      vehicleInput += rightMotor; 
      thisNode.sendOverAllFreqs(car_nodeId, vehicleInput);
    } else {
      String w = "M0a0";
      thisNode.sendOverAllFreqs(car_nodeId, w);
    }
  }

  bool notDone2 = cam.exe_apriltag_detection(targetID2, &target_cx, &target_cy, &target_angle);
  if (notDone2 == false){
    String w = "M0a0";
    thisNode.sendOverAllFreqs(car2_nodeId, w);
    if (completionCounter2 > 0){
          completionCounter2--;
    }
  } else {
    completionCounter2 = 20;
    int x_vec = 0;
    int y_vec = 0;
    if (cam.exe_apriltag_detection(vehicleID2, &vehicle_cx, &vehicle_cy, &vehicle_angle)){
      cartesianVectorGenerator(vehicle_cx, vehicle_cy, target_cx , target_cy, vehicle_angle, x_vec, y_vec);
      int leftMotor = 0;
      int rightMotor = 0;
//    Serial.print("x vec: ");
//    Serial.println(x_vec);
//    Serial.print("y vec: ");
//    Serial.println(y_vec);
      actuatorOutputGenerator(x_vec, y_vec, leftMotor, rightMotor);
      String vehicleInput = "M";
      vehicleInput += leftMotor; 
      vehicleInput += "a";
      vehicleInput += rightMotor; 
      thisNode.sendOverAllFreqs(car2_nodeId, vehicleInput);
    } else {
      String w = "M0a0";
      thisNode.sendOverAllFreqs(car2_nodeId, w);
    }
  }
}

void moveVehicle(int targetAT, int vehicleAT){
  
}

//In most cases, the vehicle and the target would not be 160 pixel away from each other. Because
//of that, we might want to scale the pixel denser while keeping the orientation information intact
void scalePixelDistance(){
  
}

//given the pixel position of the vehicle and the target's pixel position, it changes the referenc axes for the target to match
//the vehicle's orientation
void cartesianVectorGenerator(int x_start, int y_start, int x_end , int y_end, int angle, int& x_vec, int& y_vec){
  int x_temp = x_end - x_start;
  int y_temp = (y_end - y_start)*-1;
  double angle_rad = (angle - 180)*PI/180.0;
  x_vec = (x_temp*cos(angle_rad) + y_temp*sin(angle_rad))*-1;
  y_vec = (y_temp*cos(angle_rad) - x_temp*sin(angle_rad))*-1;
}

//This function basically maps cartesian plan coordinate to PWM input for differential actuator vehicles
//Considering that the vehicle's current position is the origin, the goal will be relative to the origin at each
//time step. 
void actuatorOutputGenerator(float x, float y, int& leftMotor, int& rightMotor){ //http://savagemakers.com/differential-drive-tank-drive-in-arduino/
  
  float rawLeft;
  float rawRight;
  float RawLeft;
  float RawRight;

  // first Compute the angle in deg
  // First hypotenuse
  float z = sqrt(x * x + y * y);

  // angle in radians
  float rad = acos(abs(x) / z);

  // Cataer for NaN values
  if (isnan(rad) == true) {
    rad = 0;
  }

  // and in degrees
  float angle = rad * 180 / PI;

  // Now angle indicates the measure of turn
  // Along a straight line, with an angle o, the turn co-efficient is same
  // this applies for angles between 0-90, with angle 0 the co-eff is -1
  // with angle 45, the co-efficient is 0 and with angle 90, it is 1

  float tcoeff = -1 + (angle / 90) * 2;
  float turn = tcoeff * abs(abs(y) - abs(x));
  turn = round(turn * 100) / 100;

  // And max of y or x is the movement
  float mov = max(abs(y), abs(x));

  // First and third quadrant
  if ((x >= 0 && y >= 0) || (x < 0 && y < 0)) {
    rawLeft = mov; rawRight = turn;
  } else {
    rawRight = mov; rawLeft = turn;
  }

  // Reverse polarity
  if (y < 0) {
    rawLeft = 0 - rawLeft;
    rawRight = 0 - rawRight;
  }

  // Update the values
  RawLeft = rawLeft;
  RawRight = rawRight;

  // Map the values onto the defined rang
  leftMotor = map(rawLeft, DDRIVE_MIN, DDRIVE_MAX, MOTOR_MIN_PWM, MOTOR_MAX_PWM);
  rightMotor = map(rawRight, DDRIVE_MIN, DDRIVE_MAX, MOTOR_MIN_PWM, MOTOR_MAX_PWM);
  Serial.print("left motor: ");
  Serial.println(leftMotor);
  Serial.print("right motor: ");
  Serial.println(rightMotor);
}


////MESH COMMUNICATION FUNCTIONS////

void interpretDashInstructions(String& msg){
  // Deserialize different dash topics as they are received
  String result;
  // If dash topic is Command
  if(thisNode.dashDeserialize(msg, "Manual", result)){
    // Interpret what command on the dash means
    char numChar = result[0];
    String commandContent = result.substring(1);
    if(numChar == 'I'){
        thisNode.interpretInstruction(0, commandContent);
        return;
    }
    int nodeToSend = numChar - '0';
    std::list<uint32_t> nodeList = thisNode.mesh.getNodeList();
    std::list<uint32_t>::iterator it = nodeList.begin();
    for(int i = 0; i < nodeToSend && it != nodeList.end(); i++){
      it++;
    }
    if(it == nodeList.end()){
      Serial.println("error, node # not in list");
      return;
    }
    thisNode.sendOverAllFreqs(*it, commandContent);
  }
  // If the dash topic is MasterJoystick
  else if(thisNode.dashDeserialize(msg, "MasterJoystick", result)){
     // Interpret what a dash MasterJoystick message means
     String joystickcontrol = "J";
     joystickcontrol += result;
     std::list<uint32_t> nodeList = thisNode.mesh.getNodeList();
     for(std::list<uint32_t>::iterator it = nodeList.begin(); it != nodeList.end(); it++){
         thisNode.sendOverAllFreqs(*it, joystickcontrol);
     }
     thisNode.interpretInstruction(0, joystickcontrol);
     return;
  }
}

void vehicleInstruction(char instructionType, String instrData){
  // Do something with the data receieved over the network
  // Based on what the instructionType is, do something different with the instrData
  // instructionType CAN NOT BE:
  // 'B', 'O', 'L', 'R', 'D'
  //  I am using those within ArnholdMesh for testing purposes
  switch(instructionType){
    case 'u':
      break;
     default:
      break;
  }
}

// Callback function needed for painlessMesh
void receivedCallback(uint32_t from, String &msg) {
  Serial.printf("startHere: Received from %u msg=%s\n", from, msg.c_str());
  thisNode.interpretInstruction(from, msg);
}

// Callback function needed for MQTT
void mqttCallback(char* topic, uint8_t* payload, unsigned int length) {
  char* cleanPayload = (char*)malloc(length+1);
  memcpy(cleanPayload, payload, length);
  cleanPayload[length] = '\0';
  String msg = String(cleanPayload); 
  free(cleanPayload);
  Serial.println(msg);
  interpretDashInstructions(msg);
}
